#ifndef _LIB_REGION_MAP_H
#define _LIB_REGION_MAP_H

#ifdef __cplusplus
extern "C" {
#endif
  
/* Includes from other libraries  */
#include "t3d/t3d.h"
#include "libimages3d.h"
#include "libdtm.h"
#include "libdtm_gdhe.h"

/* Structure propres  la lib */

#define RMAP_MAX_LINES		(1000)
#define RMAP_MAX_COLUMNS	(1000)
#define RMAP_NB_POSITIONS_MAX	(1000)
  
#define RMAP_ASCII_FF "REGION_MAP_ascii_file_format"
  
/* Minimum num of points required to assess a partial probability */
#define RMAP_NB_POINTS_MIN_PER_SQUARE_METER	(20) 

typedef enum rmap_classification_type {CLASSIFICATION_3D,
				       CLASSIFICATION_TEXTURE,
				       NB_CLASSIFICATION_TYPE}
  RMAP_CLASSIFICATION_TYPE;
/* The following enums are valid for any type of classification (3D, texture,
   ...)  */
typedef enum rmap_attributes_state {NOT_COMPUTED,
				    COMPUTED,
				    MINMAX_NORMALIZED,
				    STATISTICALLY_NORMALIZED}
  RMAP_ATTRIBUTES_STATE;
typedef enum rmap_probas_state {RMAP_PROBAS_NOT_COMPUTED,
				RMAP_PROBAS_COMPUTED}
  RMAP_PROBAS_STATE;
typedef enum rmap_classification_state {NOT_CLASSIFIED,
					USER_CLASSIFIED,
					EXTERNALLY_CLASSIFIED,
					SYSTEM_CLASSIFIED} 
  RMAP_CLASSIFICATION_STATE;
typedef enum rmap_fusion_type {RMAP_FUSION_OVERRIDE,
			       RMAP_FUSION_MERGE}
  RMAP_FUSION_TYPE;
  
/***********************************************************************
 *
 * 'RMAP_3D_INFOS' : all the informations derived from the 3D points that
 * projects onto a region. This is mainly the feature vector that will be used
 * to asses the geometric class of the region.
 *
 ***********************************************************************/
/* Names of the attributes computed from the 3D points */
typedef enum rmap_3d_attributes {NB_PTS,
				 MEAN_H_DISTANCE, 
				 MIN_Z, MAX_Z, 
				 DELTA_Z, 
				 MEAN_Z, 
				 VAR_Z, 
				 NX, NY, NZ,
				 VAR_NX, VAR_NY, VAR_NZ, 
				 NORM_VAR_N,
				 NB_3D_ATTRIBUTES} 
  RMAP_3D_ATTRIBUTES; 
  
/* Names of the 3D terrain classes */
typedef enum rmap_3d_class {NO_3D_CLASS,
			    RMAP_FLAT, 
			    RMAP_OBSTACLE, 
			    RMAP_ROUGH, 
			    RMAP_SLOPE,
			    NB_3D_CLASSES} 
  RMAP_3D_CLASS; 
  
  /* The 3D informations */
typedef struct rmap_3d_infos
{
  /* The 3D features computed on the region */
  RMAP_ATTRIBUTES_STATE	attributesState;
  RMAP_PROBAS_STATE	probaState;
  RMAP_CLASSIFICATION_STATE	classificationState;
  int		nbNormales; /* Nb of points with a normal vector that have been
			       added */ 
  double	attributes[NB_3D_ATTRIBUTES]; /* The computed attributes */ 
  double	newProbas[NB_3D_CLASSES];
  double	mergedProbas[NB_3D_CLASSES];
  double	growedProbas[NB_3D_CLASSES];
  RMAP_3D_CLASS label;
} RMAP_3D_INFOS;
  
/***********************************************************************
 *
 * 'RMAP_TEXTURE_INFOS' : all the texture informations of a region. This is
 * mainly the feature vector that will be used to asses the "visual class" of
 * the region.
 *
 ***********************************************************************/
/* Names of the texture attributes */
typedef enum rmap_texture_attributes {ENERGY,
				      NB_TEXTURE_ATTRIBUTES} 
  RMAP_TEXTURE_ATTRIBUTES; 
/* Names of the visual terrain classes */
typedef enum rmap_texture_class {NO_TEXTURE_CLASS,
				 GRASS, 
				 CONCRETE, 
				 GRAVEL, 
				 TARMAC,
				 NB_TEXTURE_CLASSES} 
  RMAP_TEXTURE_CLASS; 

/* The texture information */
typedef struct rmap_texture_infos
{
  /* An enum in which are stored various informations */
  RMAP_ATTRIBUTES_STATE	attributesState;
  double	attributes[NB_TEXTURE_ATTRIBUTES];
  RMAP_CLASSIFICATION_STATE	classificationState;
  double	newProbas[NB_TEXTURE_CLASSES];
  double	mergedProbas[NB_TEXTURE_CLASSES];
  RMAP_TEXTURE_CLASS label;
} RMAP_TEXTURE_INFOS;
  
/***********************************************************************
 *
 * 'RMAP_REGION' : a (square) region derived from the TM_BITMAP
 *
 ***********************************************************************/
typedef enum rmap_updateState {
   RMAP_REGION_NEVER_PERCEIVED,
   RMAP_REGION_NOT_ENOUGH_POINTS,
   RMAP_REGION_INITIALLY_PERCEIVED,
   RMAP_REGION_PERCEIVED,
   RMAP_REGION_NEWLY_PERCEIVED
} RMAP_UPDATE_STATE;

typedef struct rmap_region
{
  /* Information about its update status */
  RMAP_UPDATE_STATE	updateState; 
  int			firstDate;   /* The first date it has been updated */
  int			lastDate;    /* The last date it has been updated */
  int			nbUpdates;   /* Nb of times it has been updated */

  char			classificationType[NB_CLASSIFICATION_TYPE];
  /* 3D attributes, probas and class */
  RMAP_3D_INFOS		infos3d;
  /* Texture attributes, probas and class */
  RMAP_TEXTURE_INFOS	infosTexture;

} RMAP_REGION;

/***********************************************************************
 *
 * 'REGION_MAP' : the region map structure
 *
 ***********************************************************************/

typedef enum {RMAP_INITIALIZED, 
	      RMAP_ATTRIBUTES_COMPUTED,
	      RMAP_PROBABILITIES_COMPUTED,
	      RMAP_LABELLED_BY_USER,
	      RMAP_AUTOMATICALLY_LABELLED} RMAP_INFOCELLS;
typedef enum {RMAP_NO_BOUNDS, 
	      RMAP_BOUNDS_UPDATED} RMAP_BOUNDINFOS;

typedef struct REGION_MAP {
  RMAP_INFOCELLS	lastCellStates;
  RMAP_BOUNDINFOS	bounds;
  dtm_geo_orientation   geoOrientation; /* Specifies the orientation of the
					    x/y axes with respect to North */

  /* General geometric informations */
  double	scale;		/* Size of the regions in meters */
  int		nblig, nbcol;	/* Nb of lines and columns of the REGION_MAP */
  int		iOrigin, jOrigin; /* Pixel coordinates of the global frame
				     origin */  
  DTM_GEODESIC_POINT geodesicOrigin; /* Geodesic coordinates of the regionMap
					 origin */
  double	size_x, size_y;	/* Size of the region map (meters) */
  T3D		origin2corner;	/* The transformation between the world frame
				   origin and the top-left-corner of the first
				   region -- indices (0,0) -- of the
				   regionMap. */

  int		last_iMin, last_iMax;
  int		last_jMin, last_jMax; /* Rectangle of the last updated regions
					 */  
  int		iMin, iMax;
  int		jMin, jMax;	/* Rectangle of all the updated regions */

  int		nbUpdates;	/* Nb of times the map has been updated */ 
  int		nbNewCells;	/* Nb of cells updated at the last update */
  int		nbUpdatedCells;	/* Total nb of updated cells */
	  
  /* Robot related information */
  double	robotRadius;	 /* The robot radius */
  int		robotCellRadius; /* The robot size in number of cells */
  T3D		positions[RMAP_NB_POSITIONS_MAX]; /* The positions from which a
						     3D image has been fused */

  /* indicates the considered 3D and visual classes (should be boolean) */    
  int		nbPointsMin;
  int		classes3d[NB_3D_CLASSES];
  int		classesTexture[NB_TEXTURE_CLASSES]; 


  /* The array of regions */
  RMAP_REGION*	regions;

}  REGION_MAP;

/***********************************************************************
 *
 * This structure is used to specify for each 3D point of a DATA_IM3D structure
 * to which region it belongs.
 *
 ***********************************************************************/
typedef struct rmap_region_indices {
  int	iRegion;
  int	jRegion;
} RMAP_REGION_INDICES;


/*****************************************************************************
 *
 * The learning structure: a set of manually classified regions ("learning
 * prototypes"), with the priori conditional probabilities
 * P(feature/class). These probabilities are actually assessed for each
 * classified prototype: on line, given a computed feature vector, its a priori
 * probabilities are set as the ones of the closest prototype.
 *
 *****************************************************************************/
#define RMAP_PROTO_SPACE_FF	"PROTO_SPACE_ascii_file_format"

/* States the state of the prototype space (denoted as "PS") */
typedef enum {NIL_PROTOS, 
	      FILLED_WITH_PROTOS,
	      PROBAS_PROTOS_OK} RMAP_PS_STATE; 

#define RMAP_MAX_NB_PROTOS 10000

typedef struct RMAP_3D_PROTO_SPACE
{
  RMAP_PS_STATE		state;
  RMAP_ATTRIBUTES_STATE norm;

  /* Numbers of the images that have been used as prototypes */
  int		nbRegionMap; 
  char		regionMapFiles[RMAP_NB_POSITIONS_MAX][256]; 

  /* Informations re. the 3D prototypes */ 
  int		nbProtos;	/* Total number of learned prototypes */
  int		nbNeighbors;	/* Nb of neighbors used to compute the probas */
  char		classes[NB_3D_CLASSES]; /* The 3D classes considered */ 
  int		nbPerClass[NB_3D_CLASSES]; /* The number of prototypes per
					      considered class */
  int		attributeMask[NB_3D_ATTRIBUTES]; /* States what attributes are
					     retained for the classification */
  double	minValues[NB_3D_ATTRIBUTES];
  double	maxValues[NB_3D_ATTRIBUTES];
  double	deltaValues[NB_3D_ATTRIBUTES]; /* Bounds and variations of the
						 attributes */
  double	mean[NB_3D_ATTRIBUTES];
  double	var[NB_3D_ATTRIBUTES]; /* Means and variances */

  /* Now the prototypes */
  RMAP_REGION	prototypes[RMAP_MAX_NB_PROTOS];
} RMAP_3D_PROTO_SPACE;



/***********************************************************************/
/* Error codes */
#define rMapError(num) (rMap_PrintError(num, #num, __FILE__, __LINE__))
#define RMAP_Id  (533)
#define RMAP_OK (0)
#define RMAP_ERROR (-1)
#define RMAP_ErrNum(mod,num) ((((mod) & 0xFFFF) << 16) | ((num) & 0xFFFF))

/* Les pbs d'allocation */
#define ERR_RMAP_ALLOC_REGION_MAP		RMAP_ErrNum(RMAP_Id, 1)
#define ERR_RMAP_MALLOC				RMAP_ErrNum(RMAP_Id, 2)

/* Les paramtres invalides */
#define ERR_RMAP_SIZE				RMAP_ErrNum(RMAP_Id, 10)
#define ERR_RMAP_SCALE				RMAP_ErrNum(RMAP_Id, 11)
#define ERR_RMAP_INDICES			RMAP_ErrNum(RMAP_Id, 12)
#define ERR_RMAP_PROBA_THRESHOLD		RMAP_ErrNum(RMAP_Id, 13)
#define ERR_RMAP_ODD_SIZE			RMAP_ErrNum(RMAP_Id, 14)
#define ERR_RMAP_REGION_MAP_SIZE		RMAP_ErrNum(RMAP_Id, 15)
#define ERR_RMAP_REGION_TYPE			RMAP_ErrNum(RMAP_Id, 16)
#define ERR_RMAP_ATTRIBUTE_NUMBER		RMAP_ErrNum(RMAP_Id, 17)
#define ERR_RMAP_PROTO_SPACE_STATUS		RMAP_ErrNum(RMAP_Id, 18)

/* Erreurs relatives  des fichiers */
#define ERR_RMAP_OPEN_FILE			RMAP_ErrNum(RMAP_Id, 20)
#define ERR_RMAP_FORMAT_ASCII_FILE		RMAP_ErrNum(RMAP_Id, 21)
#define ERR_RMAP_READ_FILE			RMAP_ErrNum(RMAP_Id, 22)

/* Erreurs relatives  l'affichage */
#define ERR_RMAP_CREATION_IMA4			RMAP_ErrNum(RMAP_Id, 30)
#define ERR_RMAP_DECODAGE_IMA4			RMAP_ErrNum(RMAP_Id, 31)
#define ERR_RMAP_CREATION_IMA2			RMAP_ErrNum(RMAP_Id, 32)
#define ERR_RMAP_DECODAGE_IMA2			RMAP_ErrNum(RMAP_Id, 33)
#define ERR_RMAP_CREATION_IMA1			RMAP_ErrNum(RMAP_Id, 34)
#define ERR_RMAP_DECODAGE_IMA1			RMAP_ErrNum(RMAP_Id, 35)
#define ERR_RMAP_NOT_OKFEN			RMAP_ErrNum(RMAP_Id, 36)

/* Les structures pas comme il le faut */
#define ERR_RMAP_IM3D				RMAP_ErrNum(RMAP_Id, 40)
#define ERR_RMAP_BAD_PARAMETER			RMAP_ErrNum(RMAP_Id, 41)

/* Erreurs de dbordement */
#define ERR_RMAP_OVER_NB_POSITIONS_MAX		RMAP_ErrNum(RMAP_Id, 50)
#define ERR_RMAP_PAVE_TROP_GRAND		RMAP_ErrNum(RMAP_Id, 51)
#define ERR_RMAP_ROBOT_OUT_OF_BITMAP		RMAP_ErrNum(RMAP_Id, 52)

/* Erreurs diverses */
#define ERR_RMAP_NO_CELLS_UPDATED		RMAP_ErrNum(RMAP_Id, 97)
#define ERR_RMAP_IMPOSSIBLE_CASE		RMAP_ErrNum(RMAP_Id, 98)
#define ERR_RMAP_NOT_IMPLEMENTED		RMAP_ErrNum(RMAP_Id, 99)


/* Function prototypes */
extern int rMap_fuseIm3d ( REGION_MAP* regionMap, DATA_IM3D* im3d, T3D* main2origin );
extern void rmap_update3dAttributes ( RMAP_3D_INFOS* infos3d, DATA_PT3D* pt3d );
extern int rMap_writeAsciiMap ( REGION_MAP* regionMap, FILE* file );
extern int rMap_readAsciiMap ( FILE* file, REGION_MAP* regionMap );
extern int rMap_writeAsciiRegion ( RMAP_REGION* region, FILE* file );
extern int rMap_readAsciiRegion ( RMAP_REGION* region, FILE* file );
extern int rMap_write3DProtoSpace (RMAP_3D_PROTO_SPACE* ps, FILE* file);
extern int rMap_read3DProtoSpace (FILE* file, RMAP_3D_PROTO_SPACE* ps);
extern int rMap_readRessacMap ( FILE* file, REGION_MAP* regionMap );
extern int rMap_readRessacMapOld ( FILE* file, REGION_MAP* regionMap );
extern int rMap_readUrusMap ( FILE* file, REGION_MAP* regionMap );
extern RMAP_3D_PROTO_SPACE* rMap_create3dProtoSpace ( void );
extern int rMap_destroy3dProtoSpace ( RMAP_3D_PROTO_SPACE* ps );
extern RMAP_3D_PROTO_SPACE* rMap_copy3dProtoSpace ( RMAP_3D_PROTO_SPACE* ps );
extern int rMap_info3dProtoSpace ( RMAP_3D_PROTO_SPACE* ps );
extern int rMap_infoOne3dProto ( RMAP_3D_PROTO_SPACE* ps, int i );
extern int rMap_addMapTo3dProtoSpace ( REGION_MAP* rm, char* fileName, RMAP_3D_PROTO_SPACE* ps );
extern REGION_MAP* rMap_createMap ( void );
extern int rMap_initMap ( REGION_MAP* regionMap, double scale, int nbX, int nbY, double xOrigin, double yOrigin, double zOrigin, double robotRadius );
extern int rMap_destroyMap ( REGION_MAP* regionMap );
extern REGION_MAP* rMap_copyMap ( REGION_MAP* regionMap );
extern int rMap_infoMap ( REGION_MAP* rm );
extern int rMap_infoRegion ( REGION_MAP* regionMap, int i, int j );
extern int rMap_printRegion3dInfos ( RMAP_REGION* region );
extern int rMap_printRegionTextureInfos ( RMAP_REGION* region );
extern int rMap_computeBounds ( REGION_MAP* regionMap );
extern int rMap_PrintError ( int num, char *errorCode, char *fileName, int lineNumber );
extern REGION_MAP* rMap_autocrop ( REGION_MAP* rMap );
extern REGION_MAP* rMap_subSample ( REGION_MAP* rMap, int factor);
extern char* rMap_return3dClassString ( RMAP_3D_CLASS i );
extern char* rMap_returnTextureClassString ( RMAP_TEXTURE_CLASS i );
extern char* rMap_return3dAttributeString ( RMAP_3D_ATTRIBUTES i );
extern char* rMap_returnTextureAttributeString ( RMAP_TEXTURE_ATTRIBUTES i );
extern void rMap_updateStates ( REGION_MAP* regionMap, RMAP_REGION* region, int iRegion, int jRegion, int currentDate );
extern void rMap_endUpdate ( REGION_MAP* regionMap );
extern int rMap_setUTMGeodesicReference ( REGION_MAP* rm, int consideredPoint, dtm_geo_orientation orientation, double easting, double northing, double alt, double nextEast );
extern int rMap_computeGrowProbas ( REGION_MAP* rm );
extern int rMap_normalize3DProtoSpace (RMAP_3D_PROTO_SPACE* ps, RMAP_ATTRIBUTES_STATE normType);
extern int rMap_normalize3DAttributes (RMAP_REGION* region, RMAP_3D_PROTO_SPACE* ps, RMAP_ATTRIBUTES_STATE normType);
extern int rMap_computeProtoSpace3DProbas (RMAP_3D_PROTO_SPACE*	ps, int nbNeighbors, int* attributeMask);
extern int rMap_compute3DProbas (RMAP_REGION* region, RMAP_3D_PROTO_SPACE* ps, int nbNeighbors, int* attributeMask);
extern double rMap_compute3DVolume (double* tab_min, double* tab_max, int* attributeMask);
extern double rMap_compute3DProtoDistance (RMAP_REGION* region1, RMAP_REGION* region2, int* attributeMask);
extern int rMap_computeMap3DProbas (REGION_MAP* rm, RMAP_3D_PROTO_SPACE* ps, RMAP_FUSION_TYPE fusionType);
extern int rMap_computeMap3DLabel (REGION_MAP* rm);


#ifdef __cplusplus
}
#endif
#endif /* _LIB_REGION_MAP_H */

