/* Generic C includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Structures, protos and erreurs of the lib */
#include "libregionMap.h"
 
/**********************************************************************
 *
 * Creation of a REGION_MAP structure
 *
 **********************************************************************/

REGION_MAP*	
rMap_createMap ()

{
  REGION_MAP*	regionMap;
  
  regionMap = (REGION_MAP*) malloc (sizeof (REGION_MAP));
  if (regionMap == NULL) {
    rMapError(ERR_RMAP_ALLOC_REGION_MAP);
    return (NULL);
  }

  return (regionMap);
}

/**********************************************************************
 *
 * Initialization of a REGION_MAP 
 * TODO: isn't zOrigin useless ?
 *
 **********************************************************************/
int
rMap_initMap (REGION_MAP*	regionMap,
	      double		scale,
	      int		nbX,
	      int		nbY,
	      double		xOrigin,
	      double		yOrigin,
	      double		zOrigin,
	      double		robotRadius)
     
{
  int		i,j;
  RMAP_REGION*	region;

  /* Parameters validity */
  if ((nbX <= 0) || (nbY <= 0)) {
    printf ("The nb of cells should be strictly > 0 !\n");
    return rMapError (ERR_RMAP_SIZE);
  }
  if (scale <= 0) {
    printf ("The scale should be strictly > 0 !\n");
    return rMapError (ERR_RMAP_SCALE);
  }

  if ((nbX > RMAP_MAX_LINES) || (nbY > RMAP_MAX_COLUMNS)) {
    printf ("Size parameters exceed the maximal size\n");
    return rMapError (ERR_RMAP_SIZE);
  }

  /* Sets everything to 0 */
  memset ( (REGION_MAP*) regionMap, 0, sizeof (REGION_MAP) );
  
  /* Sets the geometric characteristics of the regionMap */
  regionMap->scale = scale;
  regionMap->nblig = nbX;
  regionMap->nbcol = nbY;
  regionMap->size_x = nbX * scale;
  regionMap->size_y = nbY * scale;

  t3dInit(&(regionMap->origin2corner), T3D_EULER, T3D_ALLOW_CONVERSION);
  regionMap->origin2corner.euler.euler[3] = xOrigin;
  regionMap->origin2corner.euler.euler[4] = yOrigin;
  regionMap->origin2corner.euler.euler[5] = zOrigin;

  /* Global frame origin coordinates in regionMap pixels */
  regionMap->iOrigin = (int) (- xOrigin / scale);
  regionMap->jOrigin = (int) (- yOrigin / scale);

  /* Initialize the bounds of the perceived area so that they will be properly
     updated */
  regionMap->iMin = regionMap->nblig;
  regionMap->iMax = 0;
  regionMap->jMin = regionMap->nbcol;
  regionMap->jMax = 0;

  regionMap->last_iMin = regionMap->nblig - 1;
  regionMap->last_iMax = 0;
  regionMap->last_jMin = regionMap->nbcol - 1;
  regionMap->last_jMax = 0;

  regionMap->nbPointsMin 
    = (int) (scale * scale * RMAP_NB_POINTS_MIN_PER_SQUARE_METER);
  regionMap->lastCellStates = RMAP_INITIALIZED;

  regionMap->robotRadius = robotRadius;
  regionMap->robotCellRadius = (int) ((robotRadius + 0.5) / scale);

  /* Allocates the regions */
  regionMap->regions = (RMAP_REGION*) malloc (nbX * nbY * sizeof(RMAP_REGION));
  if (regionMap->regions == NULL) {
    printf ("Could not allocate the array of regions !\n");
    return rMapError(ERR_RMAP_ALLOC_REGION_MAP);
  }

  /* Sets all the probabilities to 0.5. TODO XXX: this actually depends on the
     number of considered classes */
  region = regionMap->regions;
  for (i = 0; i < regionMap->nblig; i++) {
    for (j = 0; j < regionMap->nbcol; j++) {
      region->infos3d.newProbas[RMAP_FLAT] = 0.5;
      region->infos3d.newProbas[RMAP_OBSTACLE] = 0.5;
      region->infos3d.mergedProbas[RMAP_FLAT] = 0.5;
      region->infos3d.mergedProbas[RMAP_OBSTACLE] = 0.5;
      region->infos3d.growedProbas[RMAP_FLAT] = 0.5;
      region->infos3d.growedProbas[RMAP_OBSTACLE] = 0.5;
      region++;
    }
  }

  return (RMAP_OK);
}


/**********************************************************************
 *
 * Destruction of a 'REGION_MAP'
 *
 **********************************************************************/

int
rMap_destroyMap (REGION_MAP*	regionMap)

{
  free (regionMap->regions);
  free (regionMap);
  
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Copy of a 'REGION_MAP' structure
 *
 **********************************************************************/

REGION_MAP*
rMap_copyMap (REGION_MAP*	regionMap)

{
  REGION_MAP*	newMap;

  newMap = rMap_createMap ();
  if (newMap == NULL)
    return (NULL);

  memcpy (newMap, regionMap, sizeof (REGION_MAP));

  return (newMap);
}


/**********************************************************************
 *
 * Prints informations on a 'REGION_MAP'
 *
 **********************************************************************/

int
rMap_infoMap (REGION_MAP*	rm)

{
  int	i;

  printf ("Informations on a REGION_MAP structure:\n");
  printf("### Geometric information:\n");
  printf ("\tThe structure contains %d x %d regions (%d in total):\n",
	  rm->nblig, rm->nbcol, 
	  rm->nblig * rm->nbcol);
  printf ("\t(it occupies %d kBytes -- %.2f mBytes)\n", 
	  (int) (rm->nblig * rm->nbcol * sizeof (RMAP_REGION)) / 1024,
	  ((float) rm->nblig * rm->nbcol * sizeof (RMAP_REGION)) / (1024 * 1024));
  printf ("\tIt covers %f x %f meters (%d x %d pixels of size %.2f m)\n",
	  rm->size_x, rm->size_y, 
	  rm->nblig, rm->nbcol, 
	  rm->scale);
  printf ("\tThe top-left corner of the first region is located at\n"
	  "\t\t(x,y,z) = (%.2f, %.2f, %.2f) m in the global frame\n",
	  rm->origin2corner.euler.euler[3],
	  rm->origin2corner.euler.euler[4],
	  rm->origin2corner.euler.euler[5]);
  printf ("\tThe pixel coordinates of the origin are: (%d, %d)\n",
	  rm->iOrigin, rm->jOrigin);
  if (rm->geoOrientation == NO_GEO_ORIENTATION)
    printf ("\tThe map is not georeferenced\n");
  else {
    printf ("\tGeodesic reference of the top-left corner of the cell (0,0):\n");
    dtm_infoGeodesicPos (&(rm->geodesicOrigin));
    printf ("\t\tThe X axis is oriented towards ");
    switch (rm->geoOrientation) {
    case GEO_NORTH: printf ("North\n"); break;
    case GEO_WEST: printf ("West\n"); break;
    case GEO_SOUTH: printf ("South\n"); break;
    case GEO_EAST: printf ("East\n"); break;
    default: printf ("nowhere !\n");
      return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
      /*     break; */
      /* Unreached */
      ;
    }
  }
  printf ("\tRobot radius: %d pixels (%.2f m)\n", 
	  rm->robotCellRadius, rm->scale * rm->robotCellRadius);

  printf ("### Information on the updates of the map:\n");
  printf ("\tNb of updates : %d\n", 
	  rm->nbUpdates);
  printf ("\tTotal number of updated cells : %d (%d during the last update)\n", 
	  rm->nbUpdatedCells, rm->nbNewCells);
  if (rm->lastCellStates != RMAP_INITIALIZED) 
    printf ("\tLast updated regions : I from %d to %d, J from %d to %d\n",
	    rm->last_iMin, rm->last_iMax,
	    rm->last_jMin, rm->last_jMax);
  if (rm->bounds == RMAP_BOUNDS_UPDATED)
    printf ("\tTotal updated regions : I from %d to %d, J from %d to %d\n",
	    rm->iMin, rm->iMax,
	    rm->jMin, rm->jMax);
  
  printf ("### Information on the classification status of the map:\n");
  printf ("\t%d points per square meter are required to classify a cell\n",
	  "\t(%d points per cell)\n",
	  RMAP_NB_POINTS_MIN_PER_SQUARE_METER, rm->nbPointsMin);
  switch (rm->lastCellStates) {
  case RMAP_INITIALIZED:
    printf ("\tCells just allocated\n");
    break;
  case RMAP_ATTRIBUTES_COMPUTED:
    printf ("\tCells attributes computed\n");
    break;
  case RMAP_PROBABILITIES_COMPUTED:
    printf ("\tCells partial probabilities computed\n");
    break;
  case RMAP_LABELLED_BY_USER:
    printf ("\tCells labelled by the user\n");
    break;
  case RMAP_AUTOMATICALLY_LABELLED:
    printf ("\tCells automatically computed\n");
    break;
  default: 
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
    /*     break; */
    /* Unreached */
  }
  printf ("\tConsidered 3D classes:\n");
  for (i = 1; i < NB_3D_CLASSES; i++) 
    if (rm->classes3d[i] != 0) 
      printf ("\t\t%s\n", rMap_return3dClassString(i));
  printf ("\tConsidered texture classes:\n");
  for (i = 1; i < NB_TEXTURE_CLASSES; i++) 
    if (rm->classesTexture[i] != 0) 
      printf ("\t\t%s\n", rMap_returnTextureClassString(i));



  return (RMAP_OK);
}


/**********************************************************************
 *
 * Prints informations on a region of a 'REGION_MAP'
 *
 **********************************************************************/

int
rMap_infoRegion (REGION_MAP*	regionMap,
		 int		i,
		 int		j)

{
  RMAP_REGION	region;

  /* Params validity check */
  if ((i < 0) || (i >= regionMap->nblig) 
      || (j < 0) || (j >= regionMap->nbcol)) {
    printf ("Enter (I,J) so that 0 <= I < %d and 0 <= J < %d !\n",
	    regionMap->nblig, regionMap->nbcol);
    return rMapError (ERR_RMAP_INDICES);
  }

  region = regionMap->regions[i * regionMap->nbcol + j];

  /* First print geometric and update informations */
  printf ("\n### Info on the region (%d, %d) of the REGION_MAP :\n", i, j);
  printf ("\tRegion size : %.2f meters\n", regionMap->scale);
  printf ("\tPosition of its center in the global frame : (%.2f, %.2f)\n",
	  (i + 0.5) * regionMap->scale +
	  regionMap->origin2corner.euler.euler[3],
	  (j + 0.5) * regionMap->scale +
	  regionMap->origin2corner.euler.euler[4]);
  if (regionMap->geoOrientation != NO_GEO_ORIENTATION) {
    DTM_GEODESIC_POINT pos;
    double xCell, yCell;

    xCell = (i + 0.5) * regionMap->scale;
    yCell = (j + 0.5) * regionMap->scale;
    switch (regionMap->geoOrientation) {
    case GEO_NORTH:
      pos.easting = - yCell + regionMap->geodesicOrigin.easting;
      pos.northing = xCell + regionMap->geodesicOrigin.northing;
      break;
    case GEO_WEST:
      pos.easting = - xCell + regionMap->geodesicOrigin.easting;
      pos.northing = - yCell + regionMap->geodesicOrigin.northing;
      break;
    case GEO_SOUTH:
      pos.easting = yCell + regionMap->geodesicOrigin.easting;
      pos.northing = - xCell + regionMap->geodesicOrigin.northing;
      break;
    case GEO_EAST:
      pos.easting = xCell + regionMap->geodesicOrigin.easting;
      pos.northing = yCell + regionMap->geodesicOrigin.northing;
      break;
    default:
      break;
    }
    pos.nextEastMeridian = regionMap->geodesicOrigin.nextEastMeridian;
    dtm_wgs84utm2geo (pos.easting, pos.northing, pos.nextEastMeridian,
		      &(pos.lat), &(pos.lon));
    dtm_infoGeodesicPos (&pos);
  }

  printf ("\tState of the region: ");
  switch (region.updateState) {
  case RMAP_REGION_NEVER_PERCEIVED: printf ("NEVER_PERCEIVED\n"); break;
  case RMAP_REGION_INITIALLY_PERCEIVED: printf ("INITIALLY_PERCEIVED\n"); break;
  case RMAP_REGION_NOT_ENOUGH_POINTS: printf ("NOT_ENOUGH_POINTS\n"); break;
  case RMAP_REGION_PERCEIVED: printf ("PERCEIVED\n"); break;
  case RMAP_REGION_NEWLY_PERCEIVED: printf ("NEWLY_PERCEIVED\n"); break;
  default: 
    return rMapError (ERR_RMAP_IMPOSSIBLE_CASE);
    /*     break; Unreached */
  }
  printf ("\tUpdated %d times (first time at %d, last time at %d)\n",
	  region.nbUpdates, region.firstDate, region.lastDate);
  
  /* The 3D informations */
  if (region.classificationType[CLASSIFICATION_3D] != 0) {
    printf ("\n3D related informations:\n");
    printf ("\tConsidered 3D classes: ");
    for (i = 1; i < NB_3D_CLASSES; i++) 
      if (regionMap->classes3d[i] != 0) 
	printf ("%s, ", rMap_return3dClassString(i));
    printf ("\n");
    if (rMap_printRegion3dInfos (&region) != RMAP_OK)
      return (RMAP_ERROR);
  }

  /* The texture informations */
  if (region.classificationType[CLASSIFICATION_TEXTURE] != 0) {
    printf ("\nTexture related informations:\n");
    printf ("\tConsidered texture classes: ");
    for (i = 1; i < NB_TEXTURE_CLASSES; i++) 
      if (regionMap->classesTexture[i] != 0) 
	printf ("%s, ", rMap_returnTextureClassString(i));
    printf ("\n");
    if (rMap_printRegionTextureInfos (&region) != RMAP_OK)
      return (RMAP_ERROR);
  }

  return (RMAP_OK);
}


/**********************************************************************
 *
 * Prints 3D informations on a 'RMAP_REGION'
 *
 **********************************************************************/

int
rMap_printRegion3dInfos (RMAP_REGION*	region)
{
  int i;

  if (region->classificationType[CLASSIFICATION_3D] != 0) {
    /* Prints the 3D attributes informations */
    switch (region->infos3d.attributesState) {
    case NOT_COMPUTED: printf ("Attributes not computed\n"); break;
    case COMPUTED: printf ("Attributes computed\n"); break;
    case MINMAX_NORMALIZED: printf ("Attributes normalized (MinMax norm)\n"); break;
    case STATISTICALLY_NORMALIZED: printf ("Attributes normalized (statistically)\n"); break;
    default:
      printf ("Impossible case !?\n");
      return (ERR_RMAP_IMPOSSIBLE_CASE);
      /* break; Unreached */
    }
    if (region->infos3d.attributesState != NOT_COMPUTED) {
      for (i = 0; i < NB_3D_ATTRIBUTES; i++) 
	printf ("\t%s: %.3f\n",
		rMap_return3dAttributeString(i), 
		region->infos3d.attributes[i]);
      printf ("\n");
    }
    /* Now the classification related information */
    switch (region->infos3d.classificationState) {
    case NOT_CLASSIFIED: 
      printf ("Not classified\n"); 
      break;
    case USER_CLASSIFIED: 
      printf ("Classified by the user as %s\n", 
	      rMap_return3dClassString (region->infos3d.label)); 
      break;
    case EXTERNALLY_CLASSIFIED: 
      printf ("Classified by an external process as %s\n",
	      rMap_return3dClassString (region->infos3d.label)); 
      break;
    case SYSTEM_CLASSIFIED: 
      printf ("Classified by the system as %s\n",
	      rMap_return3dClassString (region->infos3d.label));
      break;
    default:
      printf ("Impossible case !?\n");
      return (ERR_RMAP_IMPOSSIBLE_CASE);
      /* break; Unreached */
    }
    /* And the partial probabilities */
    switch (region->infos3d.probaState) {
    case RMAP_PROBAS_NOT_COMPUTED:
      printf ("Partial probabilities not computed\n");
      break;
    case RMAP_PROBAS_COMPUTED:
      printf ("Partial probabilities computed:\n");
      for (i = 1; i < NB_3D_CLASSES; i++) {
	printf ("\t%s : new: %.2f, merged: %.2f, growed: %.2f\n", 
		rMap_return3dClassString (i), 
		region->infos3d.newProbas[i],
		region->infos3d.mergedProbas[i],
		region->infos3d.growedProbas[i]);
      }
      break;
    default:
      printf ("Impossible case !?\n");
      return (ERR_RMAP_IMPOSSIBLE_CASE);
      /* break; Unreached */
    }

  }
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Prints texture informations on a 'RMAP_REGION'
 *
 **********************************************************************/

int
rMap_printRegionTextureInfos (RMAP_REGION*	region)
{
  if (region->classificationType[CLASSIFICATION_TEXTURE] != 0) {
    /* Prints the texture attributes informations */
    switch (region->infosTexture.attributesState) {
    case NOT_COMPUTED: printf ("Attributes not computed\n"); break;
    case COMPUTED: printf ("Attributes just computed\n"); break;
    case MINMAX_NORMALIZED: printf ("Attributes normalized (MinMax norm)\n"); break;
    case STATISTICALLY_NORMALIZED: printf ("Attributes normalized (statistically)\n"); break;
    default:
      printf ("Impossible case !?\n");
      return (ERR_RMAP_IMPOSSIBLE_CASE);
      /* break; Unreached */
    }
    if (region->infosTexture.attributesState != NOT_COMPUTED) {
      printf ("Soon, the texture attributes...\n");
    }
    /* Now the classification related information */
    switch (region->infosTexture.classificationState) {
    case NOT_CLASSIFIED: printf ("Not classified\n"); break;
    case USER_CLASSIFIED: printf ("Classified by the user\n"); break;
    case SYSTEM_CLASSIFIED: printf ("Classified by the system)\n"); break;
    default:
      printf ("Impossible case !?\n");
      return (ERR_RMAP_IMPOSSIBLE_CASE);
      /* break; Unreached */
    }
  }
  return (RMAP_OK);
}

/**********************************************************************
 *
 * Normalize the 3D attributes of a RMAP_REGION
 *
 **********************************************************************/

int
rMap_normalize3DAttributes (RMAP_REGION*		region,
			    RMAP_3D_PROTO_SPACE*	ps,
			    RMAP_ATTRIBUTES_STATE	normType)
{
  int j;

  switch (normType) {
  case MINMAX_NORMALIZED:
    for (j = 0; j < NB_3D_ATTRIBUTES; j++) {
      region->infos3d.attributes[j] 
	= (region->infos3d.attributes[j] - ps->minValues[j]) /
	ps->deltaValues[j];
    }
    break;
  case STATISTICALLY_NORMALIZED:
    for (j = 0; j < NB_3D_ATTRIBUTES; j++) {
      region->infos3d.attributes[j] 
	= (region->infos3d.attributes[j] - ps->mean[j]) / 
	sqrt (ps->var[j]);
    }
    break;
  default:
    return rMapError (ERR_RMAP_BAD_PARAMETER);
    /*       break; */
  }
  
  region->infos3d.attributesState = normType;

  return (RMAP_OK);
}
