/* Generic C includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Structures, protos and erreurs of the lib */
#include "libregionMap.h"

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

 
/**********************************************************************
 *
 * Computes the min/max coordinates of updated cells (this function should
 * actually never be called, as the bounds informations are properly updated
 * when fusing 3D images in the regionMap).
 *
 **********************************************************************/

int
rMap_computeBounds (REGION_MAP*	regionMap)

{
  RMAP_REGION*	region;
  int		i,j;

  region = regionMap->regions;
  for (i = 0; i < regionMap->nblig; i++) {
    for (j = 0; j < regionMap->nbcol; j++) {
      if (region->updateState != RMAP_REGION_NEVER_PERCEIVED) {
	regionMap->iMin = MIN (regionMap->iMin, i);
	regionMap->iMax = MAX (regionMap->iMax, i);
	regionMap->jMin = MIN (regionMap->jMin, j);
	regionMap->jMax = MAX (regionMap->jMax, j);
      }
      region++;
    }
  }

  regionMap->bounds = RMAP_BOUNDS_UPDATED;
  return (RMAP_OK);
}

/*
 * Simple function called by the macro "tbError"
 * Prints infos on the error that occured
 */

int
rMap_PrintError(int num, 
		char *errorCode, 
		char *fileName, 
		int lineNumber)
{
  printf ("ErrorCode: %s\n", errorCode);
  printf ("File: %s\n", fileName);
  printf ("Line: %d\n", lineNumber);
  return (num);
}

/**********************************************************************
 *
 * Creates a new REGION_MAP whose size strictly fits the bounding areas of the
 * given REGION_MAP
 *
 **********************************************************************/

REGION_MAP*
rMap_autocrop (REGION_MAP*     rMap)
{
  REGION_MAP*	new_rMap;
  RMAP_REGION*	new_region;

  int		nblig, nbcol;

  int		i_min, j_min;
  int		new_nblig, new_nbcol;
  double	xOrigin, yOrigin, zOrigin;
  
  int		i, j;
  
  nblig = rMap->nblig;
  nbcol = rMap->nbcol;
  
  /* Compute the bounds of the rMap */
  rMap_computeBounds (rMap);
  
  /* Compute various parameters that specifies the new one */
  i_min = rMap->iMin;
  j_min = rMap->jMin;
  new_nblig = rMap->iMax - rMap->iMin + 1;
  new_nbcol = rMap->jMax - rMap->jMin + 1;

  /* Finds the metric coordinates of the first region */
  xOrigin = i_min * rMap->scale + rMap->origin2corner.euler.euler[3];
  yOrigin = j_min * rMap->scale + rMap->origin2corner.euler.euler[4];
  zOrigin = rMap->origin2corner.euler.euler[5];
  
  /* Creates and initialize the new rMap */
  new_rMap = rMap_createMap ();
  if (new_rMap == NULL) {
    rMapError(ERR_RMAP_ALLOC_REGION_MAP);
    return (NULL);
  }
  
  if (rMap_initMap (new_rMap, rMap->scale,
		    new_nblig, new_nbcol, 
		    xOrigin, yOrigin, zOrigin,
		    rMap->robotRadius) != RMAP_OK) {
    rMapError (ERR_RMAP_ALLOC_REGION_MAP);
    return (NULL);
  }
  
  /* Copy some general update informations */
  new_rMap->lastCellStates = rMap->lastCellStates;
  new_rMap->bounds = rMap->bounds;
  new_rMap->geoOrientation = rMap->geoOrientation;
  new_rMap->nbUpdates = rMap->nbUpdates;
  new_rMap->nbNewCells = rMap->nbNewCells;
  new_rMap->nbUpdatedCells = rMap->nbUpdatedCells;
  memcpy (&(new_rMap->geodesicOrigin), &(rMap->geodesicOrigin), 
	  sizeof (DTM_GEODESIC_POINT));
  for (i = 0; i < RMAP_NB_POSITIONS_MAX; i++) 
    new_rMap->positions[i] = rMap->positions[i];
  new_rMap->nbPointsMin = rMap->nbPointsMin;
  memcpy (new_rMap->classes3d, rMap->classes3d, 
	  NB_3D_CLASSES * sizeof (int));
  memcpy (new_rMap->classesTexture, rMap->classesTexture, 
	  NB_3D_CLASSES * sizeof (int));

  /* Now loop on the regions and copy them */
  new_region = new_rMap->regions;
  for (i = 0; i < new_rMap->nblig; i++) {
    for (j = 0; j < new_rMap->nbcol; j++) {
      memcpy ((RMAP_REGION*) new_region, 
	      (RMAP_REGION*) (rMap->regions + nbcol * (i + i_min) + (j + j_min)),
	      sizeof (RMAP_REGION));
      new_region++;
    }
  }
  
  
  return (new_rMap);
}

/**********************************************************************
 *
 * Creates a new REGION_MAP which is subsampled by a intiger factor
 *
 **********************************************************************/

REGION_MAP*
rMap_subSample (REGION_MAP*     rMap,
		int		factor)
{
  REGION_MAP*	new_rMap;
  RMAP_REGION*	new_region;
  RMAP_REGION*	region;
  RMAP_REGION*	mostObstacleMergedRegion;

  int		nblig, nbcol;
  int		new_nblig, new_nbcol;
  double	xOrigin, yOrigin, zOrigin;
  
  int		i, j, k, l;
  double	mergedProbaObstMax;
  
  nblig = rMap->nblig;
  nbcol = rMap->nbcol;

  new_nblig = nblig / factor;
  new_nbcol = nbcol / factor;

  /* Creates and initialize the new rMap */
  new_rMap = rMap_createMap ();
  if (new_rMap == NULL) {
    rMapError(ERR_RMAP_ALLOC_REGION_MAP);
    return (NULL);
  }
  
  /* Finds the metric coordinates of the first region */
  xOrigin = rMap->origin2corner.euler.euler[3];
  yOrigin = rMap->origin2corner.euler.euler[4];
  zOrigin = rMap->origin2corner.euler.euler[5];

  if (rMap_initMap (new_rMap, rMap->scale * factor,
		    new_nblig, new_nbcol, 
		    xOrigin, yOrigin, zOrigin,
		    rMap->robotRadius) != RMAP_OK) {
    rMapError (ERR_RMAP_ALLOC_REGION_MAP);
    return (NULL);
  }

  /* Now loop on the regions and copy them */
  new_region = new_rMap->regions;
  for (i = 0; i < new_nblig; i++) {
    for (j = 0; j < new_nbcol; j++) {
      region 
	= rMap->regions + (nbcol * i * factor)  + (j * factor);
      mostObstacleMergedRegion = region;
      mergedProbaObstMax = 0.0;
      for (k = 0; k < factor; k++) {
	for (l = 0; l < factor; l++) {
	  region 
	    = rMap->regions + (nbcol * ((i * factor) + k))  + ((j * factor) + l);
	  if (region->infos3d.mergedProbas[RMAP_OBSTACLE] > mergedProbaObstMax) {
	    mergedProbaObstMax = region->infos3d.mergedProbas[RMAP_OBSTACLE];
	    mostObstacleMergedRegion = region;
	  }
	}
      }
      memcpy ((RMAP_REGION*) new_region, 
	      (RMAP_REGION*) mostObstacleMergedRegion,
	      sizeof (RMAP_REGION));
      new_region++;
    }
  }

  /* Computes some general attributes */
  rMap_computeBounds (new_rMap);
  new_rMap->lastCellStates = rMap->lastCellStates;
  new_rMap->nbUpdates = rMap->nbUpdates;
  
  new_region = new_rMap->regions;
  for (i = 0; i < new_nblig; i++) {
    for (j = 0; j < new_nbcol; j++) {
      if (new_region->updateState != RMAP_REGION_NEVER_PERCEIVED) 
	new_rMap->nbUpdatedCells++;
      if (new_region->lastDate == new_rMap->nbUpdates) {
	new_rMap->nbNewCells += 1;
	new_rMap->last_iMin = MIN (new_rMap->last_iMin, i);
	new_rMap->last_iMax = MAX (new_rMap->last_iMax, i);
	new_rMap->last_jMin = MIN (new_rMap->last_jMin, j);
	new_rMap->last_jMax = MAX (new_rMap->last_jMax, j);
      }
      new_region++;
    }
  }

  for (i = 0; i < RMAP_NB_POSITIONS_MAX; i++) 
    new_rMap->positions[i] = rMap->positions[i];
  
  return (new_rMap);
}

/**************************************************************
 *
 * Simply returns the string that correspond to a 3D class
 *
 **************************************************************/
char* 
rMap_return3dClassString (RMAP_3D_CLASS i)
{
  char*  classString;
  switch (i) {
  case NO_3D_CLASS: classString = "NO_3D_CLASS"; break;
  case RMAP_FLAT: classString = "RMAP_FLAT"; break;
  case RMAP_OBSTACLE: classString = "RMAP_OBSTACLE"; break;
  case RMAP_ROUGH: classString = "RMAP_ROUGH"; break;
  case RMAP_SLOPE: classString = "RMAP_SLOPE"; break;
  default: classString = "Undefined !!!"; break;
  }
  return (classString);
}

/**************************************************************
 *
 * Simply returns the string that correspond to a visual class
 *
 **************************************************************/
char* 
rMap_returnTextureClassString (RMAP_TEXTURE_CLASS i)
{
  char*  classString;
  switch (i) {
  case NO_TEXTURE_CLASS: classString = "NO_TEXTURE_CLASS"; break;
  case GRASS: classString = "GRASS"; break;
  case CONCRETE: classString = "CONCRETE"; break;
  case GRAVEL: classString = "GRAVEL"; break;
  case TARMAC: classString = "TARMAC"; break;
  default: classString = "Undefined !!!"; break;
  }
  return (classString);
}

/**************************************************************
 *
 * Simply returns the string that correspond to a 3D attribute
 *
 **************************************************************/
char* 
rMap_return3dAttributeString (RMAP_3D_ATTRIBUTES i)
{
  char*  classString;
  switch (i) {
  case NB_PTS: classString = "NB_PTS"; break;
  case MEAN_H_DISTANCE: classString = "MEAN_H_DISTANCE"; break;
  case MIN_Z: classString = "MIN_Z"; break;
  case MAX_Z: classString = "MAX_Z"; break;
  case DELTA_Z: classString = "DELTA_Z"; break;
  case MEAN_Z: classString = "MEAN_Z"; break;
  case VAR_Z: classString = "VAR_Z"; break;
  case NX: classString = "NX"; break;
  case NY: classString = "NY"; break;
  case NZ: classString = "NZ"; break;
  case VAR_NX: classString = "VAR_NX"; break;
  case VAR_NY: classString = "VAR_NY"; break;
  case VAR_NZ: classString = "VAR_NZ"; break;
  case NORM_VAR_N: classString = "NORM_VAR_N"; break;
  default: classString = "Undefined !!!"; break;
  }
  return (classString);
}

/**************************************************************
 *
 * Simply returns the string that correspond to a texture attribute
 *
 **************************************************************/
char* 
rMap_returnTextureAttributeString (RMAP_TEXTURE_ATTRIBUTES i)
{
  char*  classString;
  switch (i) {
  case ENERGY: classString = "ENERGY"; break;
  default: classString = "Undefined !!!"; break;
  }
  return (classString);
}


/**************************************************************
 *
 * Updates some region and regionMap parameters when a region is updated for 
 * first time.
 *
 **************************************************************/
void
rMap_updateStates (REGION_MAP*	regionMap,
		   RMAP_REGION*	region,
		   int		iRegion,
		   int		jRegion,
		   int		currentDate)
  
{
  region->updateState = RMAP_REGION_NEWLY_PERCEIVED;
  region->classificationType[CLASSIFICATION_3D] = 1;
  region->nbUpdates += 1;
  if (region->lastDate == 0) /* First time we perceive it */
    region->firstDate = currentDate;
  region->lastDate = currentDate;
  regionMap->nbNewCells += 1;
  regionMap->last_iMin = MIN (regionMap->last_iMin, iRegion);
  regionMap->last_iMax = MAX (regionMap->last_iMax, iRegion);
  regionMap->last_jMin = MIN (regionMap->last_jMin, jRegion);
  regionMap->last_jMax = MAX (regionMap->last_jMax, jRegion);
  return;
}

/**************************************************************
 *
 * Update the various global informations` on the regionMap after some of its
 * regions have been updated
 *
 **************************************************************/
void
rMap_endUpdate (REGION_MAP*	regionMap)

{
  regionMap->iMin = MIN (regionMap->last_iMin, regionMap->iMin);
  regionMap->iMax = MAX (regionMap->last_iMax, regionMap->iMax);
  regionMap->jMin = MIN (regionMap->last_jMin, regionMap->jMin);
  regionMap->jMax = MAX (regionMap->last_jMax, regionMap->jMax);
  regionMap->nbUpdates += 1;
  regionMap->nbUpdatedCells += regionMap->nbNewCells;
  regionMap->bounds = RMAP_BOUNDS_UPDATED;
  regionMap->lastCellStates = RMAP_ATTRIBUTES_COMPUTED;
  return;
}

/**********************************************************************
 *
 * Sets the geodesic reference of the regionMap (of the top-left corner of the
 * pixel (0,0) of the DTM.
 *
 **********************************************************************/

int 
rMap_setUTMGeodesicReference (REGION_MAP*	 rm,
			     int                 consideredPoint,
			     dtm_geo_orientation orientation,
			     double              easting,
			     double              northing,
			     double              alt,
			     double              nextEast)
{
  if ((orientation < 1) || (orientation > 4))
    return rMapError (ERR_RMAP_BAD_PARAMETER);

  if (consideredPoint != 0) {
    /* The provided geodetic coordinates are provided for the frame origin of
       the REGIONMAP */  
    switch (orientation) {
    case 1:			/* X = North */
      rm->geodesicOrigin.easting = easting - rm->origin2corner.euler.euler[4];
      rm->geodesicOrigin.northing = northing + rm->origin2corner.euler.euler[3];
      break;
    case 2:			/* X = West */
      rm->geodesicOrigin.easting = easting - rm->origin2corner.euler.euler[3];
      rm->geodesicOrigin.northing = northing - rm->origin2corner.euler.euler[4];
      break;
    case 3:			/* X = South */
      rm->geodesicOrigin.easting = easting + rm->origin2corner.euler.euler[4];
      rm->geodesicOrigin.northing = northing - rm->origin2corner.euler.euler[3];
      break;
    case 4:			/* X = East (should usually be the chosen, to
				   fit the ENU standard) */
      rm->geodesicOrigin.easting = easting + rm->origin2corner.euler.euler[3];
      rm->geodesicOrigin.northing = northing + rm->origin2corner.euler.euler[4];
      break;
    default:
      break; 
    }
  } else {
    /* The provided geodetic coordinates are the ones of the "top-left" pixel of
       the REGIONMAP, i.e. indices (0,0) */ 
    rm->geodesicOrigin.easting = easting;
    rm->geodesicOrigin.northing = northing;
  }
  rm->geodesicOrigin.alt = alt;
  rm->geoOrientation = orientation;
  rm->geodesicOrigin.nextEastMeridian = DEG_TO_RAD(nextEast);
  
  /* Computes the corresponding LLA coordinates */
  dtm_wgs84utm2geo (rm->geodesicOrigin.easting,
		    rm->geodesicOrigin.northing, 
		    rm->geodesicOrigin.nextEastMeridian,
		    &(rm->geodesicOrigin.lat), 
		    &(rm->geodesicOrigin.lon));

  return (RMAP_OK);
}

/**********************************************************************
 *
 * Computes the growedProbas
 *
 ***********************************************************************/

int
rMap_computeGrowProbas (REGION_MAP*	rm)

{
  RMAP_REGION*	region;
  int		i,j, k, l;
  double	maxObstacleProba;
  int		size = rm->robotCellRadius;

  for (i = MAX (rm->last_iMin - size, size); 
       i < MIN ((rm->nblig - size - 1), rm->last_iMax + size); 
       i++) {
    for (j = MAX (rm->last_jMin - size, size);
	 j < MIN ((rm->nbcol - size - 1), rm->last_jMax + size); 
	 j++) {

      maxObstacleProba = 0.0;
      for (k = i - size; k <= i + size; k++) {
	for (l = j - size; l <= j + size; l++) {
	  region = &(rm->regions[k * rm->nbcol + l]);
	  if (sqrt ((double) (k - i)*(k - i) + (l - j)*(l - j)) 
	      <= (double) (size)) 
	    maxObstacleProba = MAX (maxObstacleProba,
				    region->infos3d.mergedProbas[RMAP_OBSTACLE]);
	}
      }
      region = &(rm->regions[i * rm->nbcol + j]);
      if (maxObstacleProba > region->infos3d.mergedProbas[RMAP_OBSTACLE]) {
	region->infos3d.growedProbas[RMAP_OBSTACLE] = maxObstacleProba;
	region->infos3d.growedProbas[RMAP_FLAT] = 1.0 - maxObstacleProba;
	region->lastDate = rm->nbUpdates;
      }
      else {
	region->infos3d.growedProbas[RMAP_OBSTACLE] 
	  = region->infos3d.mergedProbas[RMAP_OBSTACLE];
	region->infos3d.growedProbas[RMAP_FLAT] 
	  = region->infos3d.mergedProbas[RMAP_FLAT];
      }
    }
  }
  
  return (RMAP_OK);
}
